import numpy as np
from unittest.mock import patch


from topos.core.faces import FaceArray


class DummyArray(FaceArray):
    """Dummy face array used for testing."""

    @property
    def num_sides(self):
        return 4

    @property
    def name(self):
        return "Dummy"


class TestInit(object):

    @patch('topos.core.faces.raiseError')
    def test_with_bad_type(self, Err):

        DummyArray([1, 2, 3, 4])
        Err.assert_called_once_with("FA01.1")

    @patch('topos.core.faces.raiseError')
    def test_with_bad_shape(self, Err):

        DummyArray(np.array([1, 2, 3, 4]))
        Err.assert_called_once_with("FA01.2", sides=4)

    @patch('topos.core.faces.raiseError')
    def test_with_bad_dtype(self, Err):

        DummyArray(np.array([[1, 2, 3., 4]]))
        Err.assert_called_once_with("FA01.3")



from unittest.mock import patch
from topos.core.codefu import prepare_vertex_array_function


class TestPrepareFVertexArray(object):

    @patch('topos.core.codefu.raiseError')
    def test_num_args_check(self, Err):

        prepare_vertex_array_function(lambda x, y, r, t: 4)
        Err.assert_called_once_with("VA02.2")



from .errors import raiseError
from .faces import FaceArray
from .vertices import VertexArray


class Mesh(object):

    def __init__(self, verts=None, faces=None, name=None):

        if verts is not None and not isinstance(verts, (VertexArray,)):
            raiseError('ME01.1')

        if faces is not None and not isinstance(faces, (FaceArray,)):
            raiseError('ME01.2')


        self._verts = verts
        self._faces = faces
        self._name = name

    def __repr__(self):
        s = "{}\n".format(self.name)

        if self.vertices is not None:
            s += "Vertices: {}\n".format(self.vertices.length)

        if self.faces is not None:
            s += "Faces: {}\n".format(self.faces.length)

        return s

    @property
    def vertices(self):
        return self._verts

    @property
    def faces(self):
        return self._faces

    @property
    def name(self):

        if self._name is None:
            return "Mesh"

        return self._name


    def save(self, filename):

        with open(filename, 'w') as f:

            f.write("o {}\n".format(self.name))

            if self.vertices is not None:
                f.write(self.vertices.fmt("v {} {} {}", suffix="\n"))

            if self.faces is not None:
                face_str = " ".join("{}" for _ in range(self.faces.num_sides))
                f.write(self.faces.fmt("f " + face_str))



fmt_cases = {
    # Args and kwargs
    (True, True): lambda s, a, kw: s.format(*a, **kw),

    # Args but no kwargs
    (True, False): lambda s, a, kw: s.format(*a),

    # Kwargs but no args
    (False, True): lambda s, a, kw: s.format(**kw),

    # Nothing
    (False, False): lambda s, a, kw: s
}


class Error(object):

    docs_url = 'https://topos.readthedocs.io/en/latest/troubleshooting.html'

    def __init__(self, err_type, msg, url=None):
        self.err_type = err_type
        self.msg = msg
        self.url = url

    def raiseError(self, err_code, *args, **kwargs):

        key = (args != (), kwargs != {})
        msg = err_code + ' ' + fmt_cases[key](self.msg, args, kwargs)

        if self.url is not None:
            msg += "\nMore info: " + self.docs_url + self.url

        raise self.err_type(msg)


ERRORS = {

    # -- Face Array Errors --
    'FA01.1': Error(TypeError, "Faces must be represented with a numpy array", "#fa01"),
    'FA01.2': Error(TypeError, "Face array must have shape (n, {sides})", "#fa01"),
    'FA01.3': Error(TypeError, "Faces can only be defined using integers.", "#fa01"),

    # -- Mesh Errors --
    'ME01.1': Error(TypeError, "Vertices must be represented by a Vertex Array", "#me01"),
    'ME01.2': Error(TypeError, "Faces must be represented by a FaceArray", "#me01"),

    # -- Vertex Array Errors --

    # __init__
    'VA01.1': Error(TypeError, 'Vertex array must be represented by a numpy array', "#va01"),
    'VA01.2': Error(TypeError, "Vertex array must have shape (n, 3)", "#va01"),

    # __add__
    'VA02.1': Error(TypeError, 'Incompatible shape {shape}, array must have shape (3,)', '#va02'),
    'VA02.2': Error(TypeError, 'Addition is not supported with type {type}', "#va02"),

    # __getitem__
    'VA03.1': Error(TypeError, 'Coordinates must be specified using an iterable', '#va03'),
    'VA03.2': Error(ValueError, 'Unknown coordinate variable {var}', "#va03"),

    # _coord_setter
    'VA04.1': Error(TypeError, "Coordinate values must be specified with a numpy array", "#va04"),
    'VA04.2': Error(TypeError, "Coordinate array must have shape ({length},)", "#va04")
}


def raiseError(err_code, *args, **kwargs):

    if err_code not in ERRORS:
        raise RuntimeError('Unknown Error code: ' + err_code)

    ERRORS[err_code].raiseError(err_code, *args, **kwargs)



import numpy as np
from abc import ABC, abstractmethod


from .errors import raiseError


class FaceArray(ABC):

    def __init__(self, data):

        # Data must be a numpy array
        if isinstance(data, (np.ndarray)):

            shape = data.shape

            # With shape (n, sides)
            if len(shape) != 2 or shape[1] != self.num_sides:
                raiseError("FA01.2", sides=self.num_sides)

            # It must have an integer type
            if not issubclass(data.dtype.type, np.integer):
                raiseError("FA01.3")

            self._data = data

        else:
            raiseError("FA01.1")

    def __repr__(self):
        s = self.name + " Array: "

        if self.length == 1:
            s += "{} face".format(self.length)

        else:
            s += "{} faces".format(self.length)

        return s

    def fmt(self, fmtstr, prefix="", suffix="", sep="\n"):
        """Return a string representation of the array according to a given
        format string compatible with Python's :py:meth:`python:str.format` syntax.

        :param fmtstr: The format string to apply to each face
        :type fmtstr: str
        :param prefix: A string to include before the face data
        :type prefix: str
        :param suffix: A string to include after the face data
        :type suffix: str
        :param sep: A string to include between each face.
        :type sep: str

        :raises KeyError: The format string cannot contain any named
        substitutions e.g. :code:`{x}`

        :return: :code:`prefix + foreach face <fmtstr + sep> + suffix`
        :rtype: str
        """

        string = prefix
        string += sep.join(fmtstr.format(*f) for f in self._data)
        string += suffix

        return string


    @property
    def length(self):
        return self._data.shape[0]

    @property
    def data(self):
        return self._data

    @property
    @abstractmethod
    def num_sides(self):
        """The number of sides in each face."""
        pass

    @property
    @abstractmethod
    def name(self):
        """The name of the type of faces this array contains."""
        pass


class Quads(FaceArray):

    @property
    def num_sides(self):
        return 4

    @property
    def name(self):
        return "Quad"


class Tris(FaceArray):

    @property
    def num_sides(self):
        return 3

    @property
    def name(self):
        return "Tri"



import numpy as np
import numpy.random as npr
from math import pi
from hypothesis.strategies import integers, floats, composite, tuples, lists

# In this file we define a number of strategies representing
# particular data types. Defining them here in one place will
# help keep ourselves consistent.

# A "real" number in the range of +- 1 million
real = floats(min_value=-1e6, max_value=1e6)

# A "radial" number - the R in polar coordinates
R = floats(min_value=0.1, max_value=1e6)

# An angle, the T in polar coordinates
T = floats(min_value=0, max_value=(2*pi - 0.0001))

# A cylindrical coordinate - for mesh gen reasons we represent them as (T, Z, R)
cylin = tuples(T, real, R)
cylindrical = lists(cylin, min_size=4, max_size=512)

cart = tuples(real, real, real)
cartesian = lists(cart, min_size=4, max_size=512)

# A "size" a non negative integer representing the length of arrays etc.
size = integers(min_value=1, max_value=512)

# A strategy representing an arbitrary array of faces
@composite
def faces(draw):
    num = draw(size)
    return npr.randint(1, 256, size=(num, 4))



from pytest import raises
from hypothesis import given
from hypothesis.strategies import text
from unittest.mock import patch
from .strategies import cartesian, faces


import numpy as np
from topos.core.mesh import Mesh
from topos.core.vertices import Cartesian


class TestInit(object):

    def test_no_args(self):

        m = Mesh()

        assert m._verts is None
        assert m._faces is None
        assert m._name is None

    @patch('topos.core.mesh.raiseError')
    def test_bad_verts(self, Err):

        Mesh(verts=2)
        Err.assert_called_once_with('ME01.1')

    @patch('topos.core.mesh.raiseError')
    def test_bad_faces(self, Err):

        Mesh(faces=2)
        Err.assert_called_once_with('ME01.2')


class TestProperties(object):

    @given(vs=cartesian)
    def test_vertices(self, vs):

        cs = Cartesian(np.array(vs))
        m = Mesh(verts=cs)

        assert m.vertices == cs



from magus import get_parameters
from .errors import raiseError


def prepare_vertex_array_function(f):
    """Given a function to apply to a :code:`VertexArray` try and
    prepare it ready to be applied to the numpy array in an efficient
    way
    """

    params = get_parameters(f)

    # It only makes sense to apply with three arguments
    if len(params) > 3:
        raiseError("VA02.2")



import numpy as np


from .mesh import Mesh
from .generators import planar_faces, planar_vertices,\
        cylindrical_faces, cylindrical_vertices
from .vertices import Cartesian, Cylindrical
from .faces import Quads


class Plane(Mesh):

    def __init__(self, N=4, xmin=-1., xmax=1., ymin=-1., ymax=1.):

        if N == 1:
            verts = Cartesian(np.array([[xmin, ymin, 0.],
                                        [xmax, ymin, 0.],
                                        [xmax, ymax, 0.],
                                        [xmin, ymax, 0.]]))
        else:
            verts = Cartesian(planar_vertices(N, xmin, xmax, ymin, ymax))

        if N == 1:
            faces = Quads(np.array([[1, 2, 3, 4]]))
        else:
            faces = Quads(planar_faces(N))

        super().__init__(name="Plane", verts=verts, faces=faces)


class UnitSquare(Plane):
    def __init__(self,  N=4):
        super().__init__(self, N, xmin=0., xmax=1., ymin=0., ymax=1.0)


class Cylinder(Mesh):

    def __init__(self, N_theta=8, N_z=6):

        verts = Cylindrical(cylindrical_vertices(N_theta, N_z))
        faces = Quads(cylindrical_faces(N_theta, N_z))

        super().__init__(name="Cylinder", verts=verts, faces=faces)



from abc import ABC, abstractmethod
import numpy as np
from magus.inspect import get_parameters


from .codefu import prepare_vertex_array_function
from .errors import raiseError



class VertexArray(ABC):

    _coords = 'xyzrt'

    def __init__(self, data):

        if isinstance(data, (np.ndarray,)):

            shape = data.shape

            if len(shape) != 2 or shape[1] != 3:
                raiseError("VA01.2")

            self._data = data
        else:
            raiseError("VA01.1")

    def __repr__(self):
        s = self.system + " Array: "

        if self.length == 1:
            s += '{} vertex'.format(self.length)
        else:
            s += '{} vertices'.format(self.length)

        return s

    def __add__(self, other):

        system = self.system.lower()

        if isinstance(other, (VertexArray,)):

            us = self.__getattribute__(system)
            vs = other.__getattribute__(system)
            verts = np.concatenate((vs, us))

            return self.fromarray(verts)

        if isinstance(other, (np.ndarray,)):

            shape = other.shape

            if shape != (3,):
                raiseError('VA02.1', shape=shape)

            vs = self.__getattribute__(system)
            return self.fromarray(vs + other)

        raiseError('VA02.2', type=type(other))

    @classmethod
    def fromarray(cls, array):
        """Given a numpy array, construct a :py:class:`VertexArray` from it."""
        return cls(array)

    @property
    def system(self):
        """Return a string representing the coorindate system the array uses."""
        return type(self).__name__

    @property
    def data(self):
        """Return the underlying numpy array representing the vertex array."""
        return self._data

    @property
    def length(self):
        """Return the length of the vertex array."""
        return self._data.shape[0]

    @property
    @abstractmethod
    def cartesian(self):
        """Return a numpy representing the vertex array
        w.r.t :term:`cartesian coordinate` s."""
        pass

    @property
    @abstractmethod
    def cylindrical(self):
        """Return a numpy array representing the vertex array
        w.r.t :term:`cylindrical coordinate` s."""
        pass

    @property
    def x(self):
        """Return a numpy array containing the :math:`x` coordinate of each vertex.
        Can also be used to modify each coordinate, see the
        `documentation <http://topos.readthedocs.io/en/latest/use/reference/vertexarray.html#x>`__
        for more details."""
        return self.cartesian[:, 0]

    @property
    def y(self):
        """Return a numpy array containing the :math:`y` coordinate of each vertex.
        Can also be used to modify each coordinate, see the
        `documentation <http://topos.readthedocs.io/en/latest/use/reference/vertexarray.html#y>`__
        for more details."""
        return self.cartesian[:, 1]

    @property
    def z(self):
        """Return a numpy array containing the :math:`z` coordinate of each vertex.
        Can also be used to modify each coordinate, see the
        `documentation <http://topos.readthedocs.io/en/latest/use/reference/vertexarray.html#z>`__
        for more details."""
        return self.cartesian[:, 2]

    @property
    def r(self):
        """Return a numpy array containing the :math:`r` coordinate of each vertex.
        Can also be used to modify each coordinate, see the
        `documentation <http://topos.readthedocs.io/en/latest/use/reference/vertexarray.html#r>`__
        for more details."""
        return self.cylindrical[:, 2]

    @property
    def t(self):
        """Return a numpy array containing the :math:`t` coordinate of each vertex.
        Can also be used to modify each coordinate, see the
        `documentation <http://topos.readthedocs.io/en/latest/use/reference/vertexarray.html#t>`__
        for more details."""
        return self.cylindrical[:, 0]

    def __getitem__(self, key):

        coords = []

        try:
            for c in key:

                if c not in self._coords:
                    raiseError("VA03.2")

                coords.append(self.__getattribute__(c))

        except TypeError:
            raiseError("VA03.1")

        return np.dstack(coords)[0]

    def _coord_set_array(self, arr, var):

        if not isinstance(arr, (np.ndarray,)):
            raiseError("VA04.1")

        if arr.shape != (self.length,):
            raiseError("VA04.2")

        self.__getattribute__("set_" + var)(arr)

    def _coord_set_fn(self, fn, var):

        coords = get_parameters(fn)
        args = self[coords]

        if len(coords) == 1:
            args.shape = (self.length,)
            vfn = np.vectorize(fn)
            cs = vfn(args)
            self._coord_set_array(cs, var)
        else:
            vfn = np.vectorize(fn)
            args = args.transpose()
            cs = vfn(*args)
            self._coord_set_array(cs, var)


    def _set_coord(self, arg, var):

        if isinstance(arg, (np.ndarray,)):
            self._coord_set_array(arg, var)
            return

        if callable(arg):
            self._coord_set_fn(arg, var)
            return

        raise TypeError

    def fmt(self, fmtstr, prefix="", suffix="", sep="\n"):
        """Return a string representation of the array according to a given
        format string compatible with Python's :py:meth:`python:str.format` syntax.

        :param fmtstr: The format string to apply to each vertex
        :type fmtstr: str
        :param prefix: A string to include before the vertex data
        :type prefix: str
        :param suffix: A string to include after the vertex data
        :type suffix: str
        :param sep: A string to include between each vertex.
        :type sep: str
        :raises KeyError: The format string cannot contain any named
                          substitutions e.g. :code:`{x}`
        :return: :code:`prefix + foreach vertex <fmtstr + sep> + suffix`
        :rtype: str
        """

        string = prefix
        string += sep.join(fmtstr.format(*v) for v in self.cartesian)
        string += suffix

        return string

    @x.setter
    def x(self, value):
        self._set_coord(value, 'x')

    @y.setter
    def y(self, value):
        self._set_coord(value, 'y')

    @z.setter
    def z(self, value):
        self._set_coord(value, 'z')

    @r.setter
    def r(self, value):
        self._set_coord(value, 'r')

    @t.setter
    def t(self, value):
        self._set_coord(value, 't')

    @abstractmethod
    def set_x(self, arr):
        pass

    @abstractmethod
    def set_y(self, arr):
        pass

    @abstractmethod
    def set_z(self, arr):
        pass

    @abstractmethod
    def set_r(self, arr):
        pass

    @abstractmethod
    def set_t(self, arr):
        pass


class Cartesian(VertexArray):
    """An implementation of a :py:class:`VertexArray` representing vertices
    in Cartesian coordinates.
    """

    def set_x(self, xs):
        self._data[:, 0] = xs

    def set_y(self, ys):
        self._data[:, 1] = ys

    def set_z(self, zs):
        self._data[:, 2] = zs

    def set_r(self, rs):
        ts = self.cylindrical[:, 0]
        xs = rs * np.cos(ts)
        ys = rs * np.sin(ts)

        self._data[:, 0] = xs
        self._data[:, 1] = ys

    def set_t(self, ts):
        rs = self.cylindrical[:, 2]
        xs = rs * np.cos(ts)
        ys = rs * np.sin(ts)

        self._data[:, 0] = xs
        self._data[:, 1] = ys

    @property
    def cartesian(self):
        return self._data

    @property
    def cylindrical(self):
        XS = self._data[:, 0]
        YS = self._data[:, 1]
        ZS = self._data[:, 2]

        RS = np.sqrt(XS*XS + YS*YS)
        TS = np.arctan2(YS, XS)

        return np.dstack([TS, ZS, RS])[0]


class Cylindrical(VertexArray):
    """An implementation of a :py:class:`VertexArray` representing vertices
    in Cylindrical coordinates.
    """

    def set_x(self, xs):
        ys = self.cartesian[:, 1]

        ts = np.arctan2(ys, xs)
        rs = xs / np.cos(ts)  # This might cause problems!

        self._data[:, 0] = ts
        self._data[:, 2] = rs

    def set_y(self, ys):
        xs = self.cartesian[:, 0]

        ts = np.arctan2(ys, xs)
        rs = ys / np.sin(ts)

        self._data[:, 2] = rs
        self._data[:, 0] = ts

    def set_z(self, zs):
        self._data[:, 1] = zs

    def set_r(self, rs):
        self._data[:, 2] = rs

    def set_t(self, ts):
        self._data[:, 0] = ts

    @property
    def cylindrical(self):
        return self._data

    @property
    def cartesian(self):
        TS = self._data[:, 0]
        ZS = self._data[:, 1]
        RS = self._data[:, 2]

        # Do the conversion
        XS = RS * np.cos(TS)
        YS = RS * np.sin(TS)

        return np.dstack([XS, YS, ZS])[0]



from pytest import raises
from hypothesis import given, assume
from unittest.mock import patch
import numpy as np
import numpy.random as npr
import numpy.testing as npt


from topos.core.vertices import VertexArray, Cartesian, Cylindrical


from .strategies import cartesian, cylindrical, size, cart


# Dummy vertex array to let us test the main base class
class DummyArray(VertexArray):

    def native(self, other=None):

        if other is not None:
            return other.data

        return self.data


    @property
    def system(self):
        return 'Dummy'

    @property
    def cartesian(self):
        return self._data

    @property
    def cylindrical(self):
        return self._data

    def set_x(self, xs):
        self._data[:, 0] = xs

    def set_y(self, ys):
        self._data[:, 1] = ys

    def set_z(self, zs):
        self._data[:, 2] = zs

    def set_r(self, rs):
        self._data[:, 2] = rs

    def set_t(self, ts):
        self._data[:, 0] = ts


class TestInit(object):

    @patch('topos.core.vertices.raiseError')
    def test_init_bad_type(self, Err):

        DummyArray("string")
        Err.assert_called_once_with("VA01.1")

    @patch('topos.core.vertices.raiseError')
    def test_init_bad_shape(self, Err):

        DummyArray(np.array([[[2, 3]]]))
        Err.assert_called_once_with("VA01.2")

    @given(vs=cartesian)
    def test_with_array(self, vs):

        verts = np.array(vs)
        vertices = DummyArray(verts)
        assert (vertices._data == verts).all()


class TestProperties(object):

    @given(length=size)
    def test_length(self, length):

        verts = npr.rand(length, 3)
        vertices = DummyArray(verts)

        assert vertices.length == length

    def test_set_length(self):

        vs = DummyArray(np.zeros((4, 3)))

        with raises(AttributeError) as err:
            vs.length = 4

        assert "can't set attribute" in str(err.value)

    @given(length=size)
    def test_repr(self, length):

        assume(length != 1)

        vs = DummyArray(np.zeros((length, 3)))
        s = 'Dummy Array: {} vertices'.format(length)

        assert s == str(vs)

    def test_repr_single(self):

        vs = DummyArray(np.zeros((1, 3)))
        s = 'Dummy Array: 1 vertex'

        assert s == str(vs)

    @given(v=cart)
    def test_fmt_single(self, v):

        arr = DummyArray(np.array([v]))
        s = arr.fmt('{} {} {}')

        assert s == '{} {} {}'.format(*v)

        s = arr.fmt('{1} {2} {0}')
        assert s == '{1} {2} {0}'.format(*v)

    @given(length=size)
    def test_fmt_many(self, length):

        arr = DummyArray(npr.rand(length, 3))
        s = arr.fmt('{} {} {}')

        segments = s.split('\n')
        assert len(segments) == length

    @given(length=size)
    def test_x(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)

        assert (VS.x == arr[:, 0]).all()

    @given(length=size)
    def test_set_x_with_array(self, length):

        arr = npr.rand(length, 3)
        xs = npr.rand(length)
        VS = DummyArray(arr)
        VS.x = xs

        assert (VS.x == xs).all()

    @given(length=size)
    def test_set_x_with_fn_in_one_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        xs = np.array(VS.x)
        VS.x = lambda x: 2*x

        assert (VS.x == 2*xs).all()

    @given(length=size)
    def test_set_x_with_fn_in_many_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        xs, ys = np.array(VS.x), np.array(VS.y)
        VS.x = lambda x, y: x + y

        assert (VS.x == (xs + ys)).all()

    @given(length=size)
    def test_y(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)

        assert (VS.y == arr[:, 1]).all()

    @given(length=size)
    def test_set_y_with_array(self, length):

        arr = npr.rand(length, 3)
        ys = npr.rand(length)
        VS = DummyArray(arr)
        VS.y = ys

        assert (VS.y == ys).all()

    @given(length=size)
    def test_set_y_with_fn_in_one_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        ys = np.array(VS.y)
        VS.y = lambda y: 2*y

        assert (VS.y == 2*ys).all()

    @given(length=size)
    def test_set_y_with_fn_in_many_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        xs, ys = np.array(VS.x), np.array(VS.y)
        VS.y = lambda x, y: x + y

        assert (VS.y == (xs + ys)).all()

    @given(length=size)
    def test_z(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)

        assert (VS.z == arr[:, 2]).all()

    @given(length=size)
    def test_set_z_with_fn_in_one_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        zs = np.array(VS.z)
        VS.z = lambda z: 2*z

        assert (VS.z == 2*zs).all()

    @given(length=size)
    def test_set_z_with_array(self, length):

        arr = npr.rand(length, 3)
        zs = npr.rand(length)
        VS = DummyArray(arr)
        VS.z = zs

        assert (VS.z == zs).all()

    @given(length=size)
    def test_set_z_with_fn_in_many_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        xs, ys = np.array(VS.x), np.array(VS.y)
        VS.z = lambda x, y: x + y

        assert (VS.z == (xs + ys)).all()

    @given(length=size)
    def test_r(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)

        assert (VS.r == arr[:, 2]).all()

    @given(length=size)
    def test_set_r_with_fn_in_one_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        rs = np.array(VS.r)
        VS.r = lambda r: 2*r

        assert (VS.r == 2*rs).all()

    @given(length=size)
    def test_set_r_with_array(self, length):

        arr = npr.rand(length, 3)
        rs = npr.rand(length)
        VS = DummyArray(arr)
        VS.r = rs

        assert (VS.r == rs).all()

    @given(length=size)
    def test_set_r_with_fn_in_many_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        xs, ys = np.array(VS.x), np.array(VS.y)
        VS.r = lambda x, y: x + y

        assert (VS.r == (xs + ys)).all()

    @given(length=size)
    def test_t(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)

        assert (VS.t == arr[:, 0]).all()

    @given(length=size)
    def test_set_t_with_fn_in_one_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        ts = np.array(VS.t)
        VS.t = lambda t: 2*t

        assert (VS.t == 2*ts).all()

    @given(length=size)
    def test_set_t_with_fn_in_many_arg(self, length):

        arr = npr.rand(length, 3)
        VS = DummyArray(arr)
        ts, ys = np.array(VS.t), np.array(VS.y)
        VS.t = lambda t, y: t + y

        assert (VS.t == (ts + ys)).all()

    @given(length=size)
    def test_set_t_with_array(self, length):

        arr = npr.rand(length, 3)
        ts = npr.rand(length)
        VS = DummyArray(arr)
        VS.t = ts

        assert (VS.t == ts).all()


class TestGetItem(object):

    @given(length=size)
    def test_get_single(self, length):

        vs = npr.rand(length, 3)
        verts = DummyArray(vs)

        arr = verts['x']
        arr.shape = (length,)

        assert (arr == vs[:, 0]).all()


    @given(length=size)
    def test_get_many(self, length):

        vs = npr.rand(length, 3)
        verts = DummyArray(vs)

        arr = verts['rzt']

        assert(arr[:, 0] == vs[:, 2]).all()
        assert(arr[:, 1] == vs[:, 2]).all()
        assert(arr[:, 2] == vs[:, 0]).all()


class TestOperations(object):

    @given(vs=cartesian, us=cartesian)
    def test_vertexarray_add(self, vs, us):

        VS = Cartesian(np.array(vs))
        US = Cartesian(np.array(us))

        TS = VS + US
        assert TS.length == (VS.length + US.length)

    @given(vs=cartesian, c=cart)
    def test_array_add(self, vs, c):

        VS = Cartesian(np.array(vs))
        C = np.array(c)
        TS = VS + C

        assert TS.length == VS.length

        QS = TS + (-C)
        assert QS.length == VS.length

        # Hypothesis and floating point numbers are evil
        diff = np.round((QS.data - VS.data), decimals=0)
        assert (diff == 0).all()




class TestCartesian(object):

    def test_system_set(self):

        c = Cartesian(np.zeros((1, 3)))
        assert c.system == 'Cartesian'

    @given(vs=cartesian)
    def test_cartesian(self, vs):

        verts = Cartesian(np.array(vs))
        assert (verts.cartesian == vs).all()


class TestCylindrical(object):

    def test_system_set(self):

        c = Cylindrical(np.zeros((1, 3)))
        assert c.system == 'Cylindrical'

    @given(vs=cylindrical)
    def test_cylindrical(self, vs):

        verts = Cylindrical(np.array(vs))
        assert (verts.cylindrical == vs).all()
